from abc import ABC, abstractmethod
from typing import List, Callable, Dict, Any

from openai._types import NOT_GIVEN, NotGiven

from sirius.ai.long_term_memory import LongTermMemory
from sirius.common import DataClass
from sirius.database import ConfigurationEnum
from sirius.exceptions import OperationNotSupportedException


class LargeLanguageModel(ConfigurationEnum):
    GPT4o: str = "gpt-4o"
    GPT4oMini: str = "gpt-4o-mini"


open_ai_large_language_model_list: List["LargeLanguageModel"] = [
    LargeLanguageModel.GPT4o,
    LargeLanguageModel.GPT4oMini
]


class Context(DataClass, ABC):
    pass

    @staticmethod
    @abstractmethod
    def get_system_context(message: str) -> "Context":
        pass

    @staticmethod
    @abstractmethod
    def get_user_context(message: str) -> "Context":
        pass

    @staticmethod
    @abstractmethod
    def get_image_from_url_context(message: str, image_url: str) -> "Context":
        pass

    @staticmethod
    @abstractmethod
    def get_image_from_path_context(message: str, image_path: str) -> "Context":
        pass


class Function(DataClass, ABC):
    name: str
    description: str
    function: Callable
    parameters: Dict[str, Any]
    function_documentation: Dict[str, Any]


class Assistant(DataClass, ABC):
    large_language_model: LargeLanguageModel
    temperature: float
    context_list: List[Context] = []
    function_list: List[Function] = []
    max_tokens: int | NotGiven = NOT_GIVEN
    instructions_long_term_memory_list: List[LongTermMemory]
    knowledge_long_term_memory_list: List[LongTermMemory]

    @staticmethod
    def get(large_language_model: LargeLanguageModel,
            temperature: float | None = 0.2,
            context_list: List[Context] | None = None,
            function_list: List[Function] | None = None,
            prompt_template: str = "You are a helpful assistant",
            instructions_long_term_memory_list: List[LongTermMemory] | None = None,
            knowledge_long_term_memory_list: List[LongTermMemory] | None = None
            ) -> "Assistant":
        context_list = [] if context_list is None else context_list
        function_list = [] if function_list is None else function_list
        instructions_long_term_memory_list = [] if instructions_long_term_memory_list is None else instructions_long_term_memory_list
        knowledge_long_term_memory_list = [] if knowledge_long_term_memory_list is None else knowledge_long_term_memory_list

        if large_language_model in open_ai_large_language_model_list:
            from sirius.ai.open_ai import OpenAIGPTAssistant, OpenAIGPTContext
            context_list.append(OpenAIGPTContext.get_system_context(prompt_template))
            return OpenAIGPTAssistant(large_language_model=large_language_model,
                                      temperature=temperature,
                                      context_list=context_list,
                                      function_list=function_list,
                                      instructions_long_term_memory_list=instructions_long_term_memory_list,
                                      knowledge_long_term_memory_list=knowledge_long_term_memory_list)

        raise OperationNotSupportedException(f"{large_language_model.value} is not yet supported")

    @abstractmethod
    async def ask(self, question: str, image_url: str | None = None, image_path: str | None = None) -> str:
        pass



import json
import os
from enum import Enum
from typing import List, Dict, Any, Optional, cast, Tuple

from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_community.document_loaders import PyPDFLoader, UnstructuredMarkdownLoader, TextLoader
from langchain_community.document_loaders.base import BaseLoader
from langchain_community.document_loaders.csv_loader import CSVLoader
from langchain_community.vectorstores import FAISS
from langchain_openai import OpenAIEmbeddings

from sirius import common
from sirius.constants import EnvironmentSecret
from sirius.database import DatabaseDocument, DatabaseFile


class LongTermMemoryDocumentType(Enum):
    TEXT: str = "TEXT"
    MARKDOWN: str = "MARKDOWN"
    PDF: str = "PDF"
    CSV: str = "CSV"


class LongTermMemoryRecollection(DatabaseDocument):
    recollection: str
    l2_distance: float

    @staticmethod
    def get(search_result: Tuple) -> "LongTermMemoryRecollection":
        return LongTermMemoryRecollection(
            recollection=search_result[0].page_content,
            l2_distance=search_result[1]
        )


class LongTermMemory(DatabaseDocument):
    source: str
    document_type: LongTermMemoryDocumentType
    chunk_size: int
    chunk_overlap: int
    size: int
    file_name: str
    description: str

    async def recollect(self, question: str, max_l2_distance: float = 0.25) -> List[LongTermMemoryRecollection]:
        recollection_list: List[LongTermMemoryRecollection] = []
        embedding: OpenAIEmbeddings = OpenAIEmbeddings(disallowed_special=(), openai_api_key=common.get_environmental_secret(EnvironmentSecret.OPEN_AI_API_KEY))  # type: ignore[call-arg]
        database_file: DatabaseFile = await DatabaseFile.get(self.file_name)
        faiss: FAISS = FAISS.deserialize_from_bytes(embeddings=embedding, serialized=database_file.data, allow_dangerous_deserialization=True)
        search_vector = await embedding.aembed_query(question)

        for search_result in await faiss.asimilarity_search_with_score_by_vector(search_vector):
            long_term_memory_recollection: LongTermMemoryRecollection = LongTermMemoryRecollection.get(search_result)
            if long_term_memory_recollection.l2_distance < max_l2_distance:
                recollection_list.append(long_term_memory_recollection)

        return recollection_list

    async def recollect_for_llm(self, question: str, max_l2_distance: float = 0.25) -> str:
        recollection_list: List[LongTermMemoryRecollection] = await self.recollect(question, max_l2_distance=max_l2_distance)
        return_data: List[Dict[str, str | float]] = []
        for recollection in recollection_list:
            return_data.append({"information_chunk": recollection.recollection,
                                "accuracy_score": 1 - recollection.l2_distance})

        return json.dumps(return_data)

    @staticmethod
    async def remember_from_url(url: str, document_type: LongTermMemoryDocumentType, description: str) -> "LongTermMemory":
        temp_file_path: str = await common.download_file_from_url(url)
        return await LongTermMemory.remember(temp_file_path, document_type, description, source=url)

    @classmethod
    async def remember(cls, file_path: str,
                       document_type: LongTermMemoryDocumentType,
                       description: str,
                       chunk_size: int = 2000,
                       chunk_overlap: int = 200,
                       is_delete_after: bool = True,
                       source: str = "", ) -> "LongTermMemory":
        file_name: str = common.get_sha256_hash(file_path)
        long_term_memory: LongTermMemory | None = await LongTermMemory.find_by_file_name(file_name)

        if long_term_memory is None:
            vector_index: bytes = LongTermMemory._get_faiss(file_path, document_type, chunk_size, chunk_overlap).serialize_to_bytes()
            metadata: Dict[str, Any] = {
                "purpose": cls.__name__,
                "source": source,
                "document_type": document_type.value,
                "chunk_size": chunk_size,
                "chunk_overlap": chunk_overlap,
                "size": len(vector_index),
                "file_name": file_name,
                "description": description
            }
            long_term_memory = LongTermMemory(**metadata)
            database_file: DatabaseFile = DatabaseFile(file_name=common.get_sha256_hash(file_path), metadata=metadata, purpose=metadata["purpose"])
            database_file.load_data(vector_index)

            await database_file.save()
            await long_term_memory.save()

        if is_delete_after:
            common.run_in_separate_thread(os.remove, file_path)

        return long_term_memory

    @staticmethod
    async def find_by_file_name(file_name: str) -> Optional["LongTermMemory"]:
        long_term_memory_list: List[LongTermMemory] = cast(List[LongTermMemory], await LongTermMemory.find_by_query(LongTermMemory.model_construct(file_name=file_name)))

        if len(long_term_memory_list) == 0:
            return None
        elif len(long_term_memory_list) == 1:
            return long_term_memory_list[0]
        else:
            raise

    @staticmethod
    def _get_faiss(file_path: str, document_type: LongTermMemoryDocumentType, chunk_size: int = 500, chunk_overlap: int = 50) -> FAISS:
        text_splitter: RecursiveCharacterTextSplitter = RecursiveCharacterTextSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)
        embedding: OpenAIEmbeddings = OpenAIEmbeddings(disallowed_special=(), openai_api_key=common.get_environmental_secret(EnvironmentSecret.OPEN_AI_API_KEY))  # type: ignore[call-arg]

        if document_type == LongTermMemoryDocumentType.MARKDOWN:
            loader: BaseLoader = UnstructuredMarkdownLoader(file_path)
        elif document_type == LongTermMemoryDocumentType.PDF:
            loader = PyPDFLoader(file_path, extract_images=True)
        elif document_type == LongTermMemoryDocumentType.CSV:
            loader = CSVLoader(file_path)
        else:
            loader = TextLoader(file_path)

        return FAISS.from_documents(text_splitter.split_documents(loader.load()), embedding)



import os
from typing import List

from langchain import hub
from langchain.agents import AgentExecutor
from langchain.agents import create_openai_functions_agent
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.tools import Tool
from langchain.tools.retriever import create_retriever_tool
from langchain_community.document_loaders import WebBaseLoader
from langchain_community.document_loaders.base import BaseLoader
from langchain_community.vectorstores import FAISS
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import SystemMessage, BaseMessage, HumanMessage, AIMessage
from langchain_openai import ChatOpenAI
from langchain_openai import OpenAIEmbeddings

from sirius import common
from sirius.common import DataClass
from sirius.constants import EnvironmentSecret
from sirius.database import ConfigurationEnum


class LargeLanguageModel(ConfigurationEnum):
    GPT4o: str = "gpt-4o"
    GPT4oMini: str = "gpt-4o-mini"


class Assistant(DataClass):
    #   TODO: Fix this
    llm: BaseChatModel | None = None
    message_list: List[BaseMessage] | None = None

    def ask(self, question: str) -> str:
        self.message_list.append(HumanMessage(content=question))

        #   TODO: type, ignore added but unsure if it should actually be ignored
        ai_message: AIMessage = self.llm(self.message_list)  #   type: ignore[assignment]
        self.message_list.append(ai_message)
        return ai_message.content  # type: ignore[return-value]

    @staticmethod
    def get(large_language_model: LargeLanguageModel, temperature: float = 0.2,
            prompt_template: str = "") -> "Assistant":
        assistant: Assistant = Assistant()
        assistant.message_list = [SystemMessage(content=prompt_template)]
        assistant.llm = ChatOpenAI(model=large_language_model.value,
                                   openai_api_key=common.get_environmental_secret(EnvironmentSecret.OPEN_AI_API_KEY),
                                   temperature=temperature)  # type: ignore[call-arg]
        return assistant


class Loader(DataClass):
    loader: BaseLoader
    name: str
    description: str

    def get_retriever_tool(self, chunk_size: int = 1000, chunk_overlap: int = 200) -> Tool:
        docs = self.loader.load()
        documents = RecursiveCharacterTextSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap).split_documents(docs)
        vector = FAISS.from_documents(documents, OpenAIEmbeddings())
        return create_retriever_tool(vector.as_retriever(), self.name, self.description)

    @staticmethod
    def get_web_loader(name: str, description: str, url: str) -> "Loader":
        return Loader(loader=WebBaseLoader(url), name=name, description=description)


class Agent(DataClass):
    #   TODO: This should not be None
    prompt_template: str | None = None
    agent_executor: AgentExecutor | None = None
    chat_history: List[BaseMessage] = []

    async def ask(self, question: str) -> str:
        reply: str = (await self.agent_executor.ainvoke({"input": question, "chat_history": self.chat_history}))["output"]
        self.chat_history = self.chat_history + [HumanMessage(content=question), AIMessage(content=reply)]
        return reply

    @staticmethod
    def get(large_language_model: LargeLanguageModel, prompt_template: str = "", temperature: float = 0.2, loader_list: List[Loader] | None = None) -> "Agent":
        loader_list = [] if loader_list is None else loader_list
        prompt = hub.pull("langchain-ai/openai-functions-template").partial(instructions=prompt_template)

        #   TODO: This needs to be a named parameter but that doesn't work for some reason
        os.environ["OPENAI_API_KEY"] = common.get_environmental_secret(EnvironmentSecret.OPEN_AI_API_KEY)
        llm: ChatOpenAI = ChatOpenAI(model=large_language_model.value,
                                     # openai_api_key=common.get_environmental_secret(EnvironmentSecret.OPEN_AI_API_KEY),
                                     temperature=temperature)

        tools: List[Tool] = [loader.get_retriever_tool() for loader in loader_list]
        openai_agent = create_openai_functions_agent(llm=llm, tools=tools, prompt=prompt)

        agent: Agent = Agent()
        agent.agent_executor = AgentExecutor(agent=openai_agent, tools=tools, verbose=True)  # type: ignore[arg-type]
        agent.prompt_template = prompt_template
        return agent



import datetime
import json
from enum import Enum
from typing import Union, cast, List, Dict, Any, Optional

import motor
from bson import ObjectId
from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorDatabase, AsyncIOMotorCollection, AsyncIOMotorGridFSBucket, AsyncIOMotorGridOutCursor
from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.database import Database

from sirius import common
from sirius.common import DataClass
from sirius.constants import EnvironmentSecret
from sirius.database.exceptions import NonUniqueResultException, DocumentNotFoundException
from sirius.exceptions import SDKClientException

client: AsyncIOMotorClient | None = None
db: AsyncIOMotorDatabase | None = None
client_sync: MongoClient | None = None
db_sync: Database | None = None
fs: AsyncIOMotorGridFSBucket | None = None
configuration_cache: Dict[str, Any] = {}


async def initialize() -> None:
    global client, db, fs
    client = motor.motor_asyncio.AsyncIOMotorClient(f"{common.get_environmental_secret(EnvironmentSecret.MONGO_DB_CONNECTION_STRING)}", uuidRepresentation="standard") if client is None else client
    db = client[common.get_environmental_secret(EnvironmentSecret.APPLICATION_NAME)] if db is None else db
    fs = AsyncIOMotorGridFSBucket(db)


def initialize_sync() -> None:
    global client_sync, db_sync
    client_sync = MongoClient(f"{common.get_environmental_secret(EnvironmentSecret.MONGO_DB_CONNECTION_STRING)}", uuidRepresentation="standard") if client_sync is None else client_sync
    db_sync = client_sync[common.get_environmental_secret(EnvironmentSecret.APPLICATION_NAME)] if db_sync is None else db_sync


async def drop_collection(collection_name: str) -> None:
    await initialize()
    await cast(AsyncIOMotorDatabase, db).drop_collection(collection_name)


def drop_collection_sync(collection_name: str) -> None:
    initialize_sync()
    db_sync.drop_collection(collection_name)


class DatabaseFile(DataClass):
    id: ObjectId | None = None
    file_name: str
    purpose: str
    metadata: Dict[str, Any]
    upload_date: datetime.datetime | None = None
    local_file_path: str | None = None

    @property
    def data(self) -> bytes:
        with open(self.local_file_path, "rb") as f:
            return f.read()

    def load_data(self, data: bytes) -> None:
        self.id = None
        self.upload_date = None
        self.local_file_path = common.get_new_temp_file_path()
        with open(self.local_file_path, "wb") as f:
            f.write(data)

    async def save(self) -> None:
        global fs

        await initialize()
        if self.local_file_path is None:
            raise SDKClientException("There is no file loaded to save")
        self.metadata = {} if self.metadata is None else self.metadata
        self.metadata["purpose"] = self.purpose

        existing_database_file: DatabaseFile | None = await DatabaseFile.find(self.file_name)
        if existing_database_file is None:
            file_id = await fs.upload_from_stream(self.file_name, self.data, metadata=self.metadata)
            self.id = file_id
        else:
            await existing_database_file.delete()
            await fs.upload_from_stream_with_id(existing_database_file.id, self.file_name, self.data, metadata=self.metadata)

    async def delete(self) -> None:
        global fs

        await initialize()
        await fs.delete(self.id)

    @staticmethod
    async def get(file_name: str) -> "DatabaseFile":
        global fs

        await initialize()
        file_path: str = common.get_new_temp_file_path()
        database_file: DatabaseFile = await DatabaseFile.get_minimal(file_name)
        stream = await fs.open_download_stream(database_file.id)
        with open(file_path, "wb") as f:
            while True:
                chunk: bytes = await stream.readchunk()
                if not chunk:
                    break
                f.write(chunk)

        database_file.local_file_path = file_path
        return database_file

    @staticmethod
    async def get_minimal(file_name: str) -> "DatabaseFile":
        global fs

        await initialize()
        cursor: AsyncIOMotorGridOutCursor = fs.find({"filename": file_name})
        file_id_list: List[Dict[str, Any]] = await cursor.to_list(length=10)

        if len(file_id_list) == 1:
            file_data: Dict[str, Any] = file_id_list[0]
            return DatabaseFile(
                id=file_data["_id"],
                file_name=file_data["filename"],
                metadata=file_data["metadata"],
                upload_date=file_data["uploadDate"],
                purpose=file_data["metadata"]["purpose"]
            )
        elif len(file_id_list) == 0:
            raise DocumentNotFoundException(f"No file named: {file_name}")
        else:
            raise NonUniqueResultException(f"More than one file with the file name: {file_name}")

    @staticmethod
    async def find(file_name: str) -> Optional["DatabaseFile"]:
        try:
            return await DatabaseFile.get_minimal(file_name)
        except DocumentNotFoundException:
            return None


class DatabaseDocument(DataClass):
    id: ObjectId | None = None
    updated_timestamp: datetime.datetime | None = None
    created_timestamp: datetime.datetime | None = None

    @classmethod
    async def _get_collection(cls) -> AsyncIOMotorCollection:
        await initialize()
        global db
        return db[cls.__name__]

    @classmethod
    def _get_collection_sync(cls) -> Collection:
        initialize_sync()
        global db_sync
        return db_sync[cls.__name__]

    async def save(self) -> None:
        collection: AsyncIOMotorCollection = await self._get_collection()

        if self.id is None:
            self.created_timestamp = datetime.datetime.now()
            object_id: ObjectId = (await collection.insert_one(json.loads(self.model_dump_json(exclude={"id"})))).inserted_id
            self.__dict__.update(self.model_dump(exclude={"id"}))
            self.id = object_id
        else:
            self.updated_timestamp = datetime.datetime.now()
            await collection.replace_one({"_id": self.id}, json.loads(self.model_dump_json(exclude={"id"})))

    def save_sync(self) -> None:
        collection: Collection = self._get_collection_sync()

        if self.id is None:
            self.created_timestamp = datetime.datetime.now()
            object_id: ObjectId = collection.insert_one(self.model_dump(exclude={"id"})).inserted_id
            self.__dict__.update(self.model_dump(exclude={"id"}))
            self.id = object_id
        else:
            self.updated_timestamp = datetime.datetime.now()
            collection.replace_one({"_id": self.id}, self.model_dump(exclude={"id"}))

    async def delete(self) -> None:
        collection: AsyncIOMotorCollection = await self._get_collection()
        await collection.delete_one({'_id': self.id})

    def delete_sync(self) -> None:
        collection: Collection = self._get_collection_sync()
        collection.delete_one({'_id': self.id})

    @classmethod
    def get_model_by_raw_data(cls, raw_data: Dict[Any, Any]) -> "DatabaseDocument":
        object_id = raw_data.pop("_id")
        queried_object: DatabaseDocument = cls(**raw_data)
        queried_object.id = object_id
        return queried_object

    @classmethod
    async def find_by_id(cls, object_id: ObjectId) -> Union["DatabaseDocument", None]:
        collection: AsyncIOMotorCollection = await cls._get_collection()
        object_model: Dict[str, Any] = await collection.find_one({'_id': object_id})
        return None if object_model is None else cls.get_model_by_raw_data(object_model)

    @classmethod
    def find_by_id_sync(cls, object_id: ObjectId) -> Union["DatabaseDocument", None]:
        collection: Collection = cls._get_collection_sync()
        object_model: Dict[str, Any] = collection.find_one({'_id': object_id})
        return None if object_model is None else cls.get_model_by_raw_data(object_model)

    @classmethod
    async def find_by_query(cls, database_document: "DatabaseDocument", query_limit: int = 100) -> List["DatabaseDocument"]:
        collection: AsyncIOMotorCollection = await cls._get_collection()
        cursor = collection.find(database_document.model_dump(exclude={"id"}, exclude_none=True))
        return [cls.get_model_by_raw_data(document) for document in await cursor.to_list(length=query_limit)]

    @classmethod
    def find_by_query_sync(cls, database_document: "DatabaseDocument", query_limit: int = 100) -> List["DatabaseDocument"]:
        collection: Collection = cls._get_collection_sync()
        cursor = collection.find(database_document.model_dump(exclude={"id"}, exclude_none=True)).limit(query_limit)
        return [cls.get_model_by_raw_data(document) for document in cursor]


class Configuration(DatabaseDocument):
    type: str
    key: str
    value: str

    @classmethod
    def find_by_query_sync(cls, configuration: "Configuration", query_limit: int = 100) -> List["Configuration"]:  # type: ignore[override]
        global configuration_cache
        if configuration.type in configuration_cache and configuration.key in configuration_cache[configuration.type]:
            return [Configuration(type=configuration.type, key=configuration.key, value=configuration_cache[configuration.type][configuration.key])]

        configuration_list: List[Configuration] = cast(List[Configuration], super().find_by_query_sync(configuration, query_limit))

        if len(configuration_list) > 1:
            raise SDKClientException(f"Duplicate configurations:\n"
                                     f"Type: {configuration.type}\n"
                                     f"Key: {configuration.key}")
        elif len(configuration_list) == 1:
            existing_configuration: Configuration = configuration_list[0]

            if existing_configuration.type in configuration_cache:
                configuration_cache[existing_configuration.type][existing_configuration.key] = existing_configuration.value
            else:
                configuration_cache[existing_configuration.type] = {existing_configuration.key: existing_configuration.value}

        return configuration_list


class ConfigurationEnum(Enum):
    default_value: Any

    def __init__(self, default_value: Any):
        self.default_value = default_value
        super().__init__()

    @property
    def value(self) -> Any:
        if common.is_ci_cd_pipeline_environment() and self.name != "TEST_KEY":
            return self.default_value

        existing_configuration_list: List[Configuration] = cast(List[Configuration], Configuration.find_by_query_sync(Configuration.model_construct(type=self.__class__.__name__, key=self.name)))

        if len(existing_configuration_list) != 0:
            return existing_configuration_list[0].value
        else:
            Configuration(type=self.__class__.__name__, key=self.name, value=self.default_value).save_sync()
            # asyncio.ensure_future(Configuration(type=self.__class__.__name__, key=self.name, value=self.default_value).save())
            return self.default_value
